#ifndef __poset_h_
#define __poset_h_

#include <map>			// hash_map would be better
#include <set>
#include <stdexcept>
#include <vector>

#include <pthread.h>

namespace mozilla {

template<typename T>
class poset {
public:
  typedef T element_type;

private:
  struct Element;
  typedef std::set<Element*> element_set_type;
  typedef std::map<element_type, Element*> map_type;
  struct Element {
    Element(element_type e)
      : mElt(e),
	mRelations(),
	mBacklinks() {
    }
    element_type mElt;
    element_set_type mRelations;
    element_set_type mBacklinks;
  };

public:
  poset() 
    : mMap() {
  }

  /**
   * Return true iff $e1 <_this e2$.
   */
  bool ordered(const element_type& e1, const element_type& e2) const {
    Element* elt1 = find(e1);
    if (!elt1)
      return false;
    Element* elt2 = find(e2);
    if (!elt2)
      return false;
    // look for elt2 in transitive closure of (elt1, <)
    return in_transitive_closure(elt2, elt1);
  }

  /**
   * Add the order $e1 <_this e2$.
   *
   * If |check| is false, you better know what you're doing!  If e1<e2 is
   * already true, then you can cause n^2 space usage.  If e2<e1 is already
   * true, you'll cause the universe to implode on itself.
   */
  void order(const element_type& e1, const element_type& e2, bool check=true) 
    throw(std::logic_error) {
    if (check) {
      if (ordered(e2, e1))
	throw std::logic_error("ordering e1<e2 when e2<e1 already exists");
      // we check this not for correctness, but to prevent O(n^2) space
      else if(ordered(e1, e2))
	return;
    }
    Element* elt1 = get(e1);
    Element* elt2 = get(e2);
    elt1->mRelations.insert(elt2);
    elt2->mBacklinks.insert(elt1);
  }

  /**
   * Remove |e| from all orders.
   */
  void remove(const element_type& e) {
    Element* elt = find(e);
    if (!elt)
      return;
    // delete (elt < *)
    for (typename element_set_type::iterator it = elt->mRelations.begin();
	 it != elt->mRelations.end();
	 ++it)
      (*it)->mBacklinks.erase(elt);
    // delete (* < elt)
    for (typename element_set_type::iterator it = elt->mBacklinks.begin();
	 it != elt->mBacklinks.end();
	 ++it)
      (*it)->mRelations.erase(elt);
    mMap.erase(e);
    delete elt;
  }

private:
  Element* find(const element_type& e) const {
    typename map_type::const_iterator it = mMap.find(e);
    return (mMap.end() == it) ? 0 : it->second;
  }

  Element* get(const element_type& e) {
    Element* elt = find(e);
    if (!elt) {
      elt = new Element(e);
      mMap[e] = elt;
    }
    return elt;
  }

  // Is |needle| in the transitive closure of (start, <)?
  bool in_transitive_closure(Element* needle, Element* start) const {
    if (start == needle
	// simple scan might be faster if relations are small
	|| start->mRelations.end() != start->mRelations.find(needle))
      return true;
    for (typename element_set_type::iterator it = start->mRelations.begin(); 
	 it != start->mRelations.end();
	 ++it)
      if (in_transitive_closure(needle, *it))
	return true;
    return false;
  }

  map_type mMap;
};

} // namespace mozilla

#endif // __poset_h_
